#pragma once

#include <cstdint>
#include <vector>
#include <queue>
#include <map>
#include <Event.h>
#include <ReWindow/types/Key.h>
#include <ReWindow/types/Cursors.h>
#include <ReWindow/types/MouseButton.h>
#include <ReWindow/types/WindowEvents.h>


namespace ReWindow {
    struct KeyboardState { std::map<Key, bool> PressedKeys; };

    class MouseState;
    class RWindow;
    class OpenGLWindow;
    class VulkanWindow;
}

class ReWindow::MouseState {
public:
    struct
    {
        bool LMB = false;
        bool RMB = false;
        bool MMB = false;
        bool SideButton1 = false;
        bool SideButton2 = false;
        bool MWheelUp = false;
        bool MWheelDown = false;
    }

    Buttons;
    std::pair<int, int> Position;
    int Wheel = 0;
public:
    [[nodiscard]] bool IsDown(const MouseButton& btn) const;
    void Set(const MouseButton& btn, bool state);
    bool& operator[](const MouseButton& btn);
};

/// RWindow is a class implementation of a platform-independent window abstraction.
/// This library also provides abstractions for user-input devices, and their interaction with the window.
class ReWindow::RWindow {
protected:
    class Platform;
    friend class Platform;
    Platform* platform;

    inline static RWindow* extant;

    int width = 640;
    int height = 480;

    bool open = false; // Is the underlying OS-Window-Handle actually open.
    bool closing = false;
    bool resizable = true;
    bool fullscreen_mode = false;
    bool focused = true;
    bool vsync = false;
    bool cursor_visible = true;
    bool cursor_focused = false;
    bool toggling_cursor_focus = false;
    bool key_repeat = false;

    float delta_time = 0.f;
    float refresh_rate = 0.f;
    unsigned long long refresh_count = 0;
    std::string title = "Redacted Window";

    std::vector<RWindowEvent> eventLog;	// history of all logged window events.
    std::queue<RWindowEvent> eventQueue; //
    MouseState currentMouse;		// purrent frame mouse state.
    MouseState previousMouse;		// previous frame mouse state

    float refresh_rate_prev_1 = 0.f;
    float refresh_rate_prev_2 = 0.f;
    float refresh_rate_prev_3 = 0.f;
    float refresh_rate_prev_4 = 0.f;
    float refresh_rate_prev_5 = 0.f;
    float avg_refresh_rate = 0.0f;

    /// Returns the most accurate and recent available mouse coordinates relative to the top left corner of the renderable area (not including the title bar).
    /// @note Call this version at most **once** per-frame. It polls the X-Window server and therefore is quite slow.
    /// @see getCursorPos();
    [[nodiscard]] std::pair<int, int> GetAccurateCursorCoordinates() const;
protected:
    void LogEvent(const RWindowEvent& e) { eventLog.push_back(e); }
    //void EnqueueEvent(const RWindowEvent& e) { eventQueue.push(e);}
    [[nodiscard]] RWindowEvent GetLastEvent() const { return eventLog.back(); }
    /// Requests the operating system to make the window fullscreen. Saves the previous window size as well.
    void Fullscreen();
    /// Requests the operating system to take the window out of fullscreen mode. Previously saved window size is restored, if possible.
    void RestoreFromFullscreen();

public:
    KeyboardState currentKeyboard;  // current frame keyboard state.
    KeyboardState previousKeyboard; // previous frame keyboard state.

    /// The default constructor does not set any members, and are left uninitialized.
    RWindow();
    /// Constructs a window by explicitly setting the title, width, height, and optionally; rendering API, fullscreen, resizable, and vsync.
    /// @param wTitle The window title text.
    /// @param wWidth
    /// @param wHeight
    /// @param wFullscreen
    /// @param wResizable
    /// @param wVsync
    explicit RWindow(const std::string& wTitle, int wWidth = 640, int wHeight = 480,
                     bool wFullscreen = false, bool wResizable = true, bool wVsync = false);

    /// The destructor deallocates and cleans up any memory used by the RWindow program.
    /// @note If the window handle is not already destroyed, it will be done here.
    virtual ~RWindow();

    static RWindow* GetExtant() { return extant; };

    /// Bindables are provided for hooking into window instances conveniently, without the need to derive and override for simple use cases.
    Event<> OnOpenEvent;
    Event<> OnClosingEvent;
    Event<RWindowEvent> OnFocusLostEvent;
    Event<RWindowEvent> OnFocusGainEvent;
    Event<float> OnRefreshEvent;
    Event<WindowResizeRequestEvent> OnResizeRequestEvent;
    Event<KeyDownEvent> OnKeyDownEvent;
    Event<KeyUpEvent> OnKeyUpEvent;
    Event<MouseMoveEvent> OnMouseMoveEvent;
    Event<MouseButtonDownEvent> OnMouseButtonDownEvent;
    Event<MouseButtonUpEvent> OnMouseButtonUpEvent;
    Event<MouseWheelEvent> OnMouseWheelEvent;

    /// These methods can also be overridden in derived classes.
    /// Called upon the window requesting to open.
    virtual void OnOpen() {}

    /// Called right before the window closes.
    virtual void OnClosing() {}

    /// Called when the window loses focus.
    virtual void OnFocusLost(const RWindowEvent& e) {}

    /// Called when the window gains focus.
    virtual void OnFocusGain(const RWindowEvent& e) {}

    /// Called when the window is 'refreshed', in other words, a render pass is completed.
    virtual void OnRefresh(float elapsed) {}

    /// Called when a resize request has succeeded.
    virtual void OnResizeSuccess() {}

    /// Called when a resize request is sent to the operating system.
    virtual bool OnResizeRequest(const WindowResizeRequestEvent& e) { return true;}

    virtual void OnKeyDown(const KeyDownEvent&) {}

    virtual void OnKeyUp(const KeyUpEvent&) {}

    virtual void OnMouseMove(const MouseMoveEvent&) {}

    virtual void OnMouseButtonDown(const MouseButtonDownEvent&) {}

    virtual void OnMouseButtonUp(const MouseButtonUpEvent&) {}

    virtual void OnMouseWheel(const MouseWheelEvent&) {}

    [[nodiscard]] int GetMouseWheelPersistent() const { return currentMouse.Wheel; }

    [[nodiscard]] bool IsOpen() const { return open; }


    /// Returns whether the window is currently visible to the user.
    // TODO On Linux, Some desktop environments don't always do this when switching workspaces.
    [[nodiscard]] bool IsVisible() const;

    [[nodiscard]] bool IsClosing() const { return closing; }

    /// Returns whether the window currently has mouse and/or keyboard focus.
    [[nodiscard]] bool IsFocused() const;

    /// Returns whether the window is currently in Fullscreen.
    // TODO: Support Fullscreen, FullscreenWindowed, and Windowed?
    [[nodiscard]] bool IsFullscreen() const;

    /// Returns whether the window can be resized.
    [[nodiscard]] bool IsResizable() const;

    /// Returns whether V-Sync is enabled.
    [[nodiscard]] bool IsVsyncEnabled() const;

    /// Returns whether the window is considered to be alive. Once dead, any logic loop should be terminated, and the cleanup procedure should run.
    [[nodiscard]] bool IsAlive() const;

    /// Returns whether the given key is currently being pressed.
    [[nodiscard]] bool IsKeyDown(Key key) const;

    /// Returns whether the given mouse button is currently being pressed.
    [[nodiscard]] bool IsMouseButtonDown(const MouseButton& button) const;

    [[nodiscard]] float GetDeltaTime() const;

    /// Returns the approximate frames-per-second using delta time.
    [[nodiscard]] float GetRefreshRate() const;

    /// Returns the number of frames ran since the windows' creation.
    [[nodiscard]] unsigned long long GetRefreshCount() const;

    /// Cleans up and closes the window object.
    void Close();

    /// Closes the window immediately, potentially without allowing finalization to occur.
    void ForceClose();

    /** Display a small window with some text and an "OK" button.
     * I tried un-defining the macro, Calling it in a lambda,
     * Calling MessageBoxA, Putting the call to the MessageBox
     * function in its own namespace and then calling that.
     * It just wouldn't let me name it MessageBox - Redacted. */

    /// @note Execution of the parent window is stopped while the message box is up.
    void DialogOK(const std::string& title, const std::string& message);

    /// Sets whether the window is fullscreen.
    void SetFullscreen(bool fs);

    /// Sets whether the window can be resized by the user.
    /// @note Resize events will still be sent if fullscreen is toggled.
    void DisableResizing();

    /// Sets the title of this window.
    void SetTitle(const std::string& title);
    [[nodiscard]] std::string GetTitle() const;

    /// Returns the horizontal length of this window, in pixels.
    [[nodiscard]] int GetWidth() const;

    /// Returns the vertical length of this window, in pixels.
    [[nodiscard]] int GetHeight() const;

    /// Tells the underlying window manager to destroy this window and drop the handle.
    /// The window, in theory, can not be re-opened after this.
    void DestroyOSWindowHandle();

    /// Reads events from the operating system, and processes them accordingly.
    /// TODO: Move out of public API, consumers should call Refresh or ideally an update() call.
    void PollEvents();

    /// Updates internal window state, similar to ManagedRefresh, but without accounting for any timekeeping. This is left up to the user.
    void Refresh();

    /// Updates the window and handles timing internally.
    void ManagedRefresh();

    /// Requests the operating system to change the window size.
    /// @param width
    /// @param height
    void SetSize(int width, int height);

    /// Requests the operating system to change the window size.
    /// @param size
    void SetSize(const std::pair<int, int>& size);

    /// Returns the position of the window's top-left corner relative to the display
    [[nodiscard]] std::pair<int, int> GetPosition() const;

    /// Returns the known size of the window, in {x,y} pixel measurement.
    [[nodiscard]] std::pair<int, int> GetSize() const;

    /// Requests the operating system to move the window to the specified coordinates on the display.
    /// @param x The horizontal screen position to place the window at.
    /// @param y The vertical screen position to place the window at.
    void SetPosition(int x, int y);

    /// Requests the operating system to move the window to the specified coordinates on the display.
    /// @param pos A std::pair<int, int> representing the x,y coordinates of the desired window destination. Fractional values are ignored.
    void SetPosition(const std::pair<int, int>& pos);

    /// Pull the window to the top, such that it is displayed on top of everything else.
    /// NOTE: The implementation is defined per-OS, and thus there is no guarantee of it always working.
    void Raise();

    /// Push the window Lower, such that it is effectively hidden behind other windows.
    /// NOTE: The implementation is defined per-OS, and thus there is no guarantee of it always working.
    void Lower();

    // TODO verify functionality if the cursor is invisible.
    void SetCursorStyle(CursorStyle style) const;

    void SetCursorCustomIcon() const;

    /// @returns Where the cursor was just before we teleported it to the center.
    /// @note You should check if our window is in focus before doing this.
    /// @note This is useful for 3D games.
    std::pair<int, int> SetCursorCenter();

    /// Set the position of the cursor relative to the top-left corner of the renderable area.
    /// @returns False if the cursor could not be guaranteed to be teleported.
    /// @note our window *must* be visible and in focus.
    /// @note Moving the cursor outside our window is unsupported.
    /// @note Doesn't update where the mouse is observed for this refresh.
    [[nodiscard]] bool SetCursorPosition(const std::pair<int, int>& position);
    [[nodiscard]] bool SetCursorPosition(int x, int y);

    /// Tells the operating system to not allow the cursor to go off our window.
    /// @note This is useful for 3D games.
    void SetCursorFocused(bool state);

    /// Hides the cursor when it's inside our window.
    /// @note This is useful for 3D games.
    void SetCursorVisible(bool cursor_enable);

    /// @returns Whether the cursor is visible.
    bool GetCursorVisible();

    /// @returns Whether the cursor is focused (cannot go off of our window).
    /// @note Delayed until the next time our window is in focus.
    bool GetCursorFocused();

    /// Returns an std::pair<int, int> representing mouse coordinates relative to the top-left corner of the window.
    /// This result is cached from the operating-system each time poll events is called.
    /// @see GetAccurateMouseCoordinates().
    [[nodiscard]] std::pair<int, int> GetCursorPosition() const;

    /// Returns the current time, represented as a high-resolution std::chrono alias.
    static std::chrono::steady_clock::time_point GetTimestamp();

    /// Computes elapsed time from a start-point and end-point.
    float ComputeElapsedFrameTimeSeconds(std::chrono::steady_clock::time_point start, std::chrono::steady_clock::time_point end);

    /// Updates internals to account for the latest calculated frame time.
    void UpdateFrameTiming(float frame_time);
protected:
    void processKeyRelease(Key key);
    void processKeyPress(Key key);
    /// @note This will be invoked **before** the window-close procedure begins.
    void processOnClose();
    /// @note This will be invoked **after** the window-open procedure completes.
    void processOnOpen();
    void processMousePress(const MouseButton& btn);
    void processMouseRelease(const MouseButton& btn);
    void processFocusIn();
    void processFocusOut();
    void processMouseMove(const std::pair<int, int>& last_pos, const std::pair<int, int>& new_pos);
    void processMouseWheel(int scrolls);
    void SetSizeWithoutEvent(const std::pair<int, int>& size);

    /// Virtual functions which *must* be overridden based on the Renderer.
public:
    virtual void SwapBuffers() = 0;

    /// Sets whether vertical sync is enabled.
    /// @note This is implemented per-OS, and as such, it simply requests the OS to do what we want. No guarantee about follow-through can be given.
    virtual void SetVsyncEnabled(bool state) = 0;

    /// Sets whether we're using key repeat.
    void SetKeyRepeatEnabled(bool state);

    /// @returns If key repeat is enabled.
    [[nodiscard]] bool IsKeyRepeat() const { return key_repeat; };

    /// Returns the name of the developer of the user's graphics driver, if it can be determined.
    virtual std::string GetGraphicsDriverVendor() = 0;

    /// This function instructs the operating system to create the actual window, and give it to us to control.
    /// Calling this function, therefore, creates the real 'window' object on the operating system.
    [[nodiscard]] virtual bool Open() = 0;

    /// Bring attention to the user, On X11 this flashes the entry for this window in the window list.
    /// @note If the window is already in focus when this is called nothing happens.
    void Flash();

    /// @returns True if we are definitely running on a software renderer.
    /// @note For some APIs this isn't typically possible.
    [[nodiscard]] virtual bool SoftwareRendered() { return false; }
};

// TODO Set flags in window constructor.
class ReWindow::OpenGLWindow : public RWindow {
protected:
    uint8_t gl_major, gl_minor;
public:
    void SwapBuffers() override;
    void SetVsyncEnabled(bool state) override;
    std::string GetGraphicsDriverVendor() override;
    [[nodiscard]] bool Open() override;
    [[nodiscard]] bool SoftwareRendered() override;
public:
    OpenGLWindow(const std::string& title, int width, int height, uint8_t gl_major, uint8_t gl_minor);
};

class ReWindow::VulkanWindow : public RWindow {
public:
    void SwapBuffers() override;
    void SetVsyncEnabled(bool state) override;
    std::string GetGraphicsDriverVendor() override;
    [[nodiscard]] bool Open() override;
    VulkanWindow(const std::string& title, int width, int height);
};