#include <Windows.h>
#include <gl/GL.h>
#include <ReWindow/types/Window.h>
#include <ReWindow/Logger.h>

class ReWindow::RWindow::Platform {
public:
    HINSTANCE hInstance;
    HWND hwnd;
    HDC hdc;
    std::pair<int, int> window_size_before_fullscreen;
    std::pair<int, int> window_position_before_fullscreen;

    static LRESULT CALLBACK WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
};

using namespace ReWindow;

void RWindow::SetSize(int newWidth, int newHeight) {
    if (!resizable)
        return;

    this->width = newWidth;
    this->height = newHeight;

    DWORD style = GetWindowLong(platform->hwnd, GWL_STYLE);
    DWORD exStyle = GetWindowLong(platform->hwnd, GWL_EXSTYLE);

    RECT rect = { 0, 0, newWidth, newHeight };
    AdjustWindowRectEx(&rect, style, FALSE, exStyle);

    int totalWidth = rect.right - rect.left;
    int totalHeight = rect.bottom - rect.top;

    SetWindowPos(platform->hwnd, nullptr, 0, 0, totalWidth, totalHeight,
                 SWP_NOMOVE | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_FRAMECHANGED);
}


void RWindow::SetCursorFocused(bool state) {
    toggling_cursor_focus = state;

    if (state)
        cursor_focused = true;

    else if (!state && cursor_focused) {
        ClipCursor(nullptr);
        cursor_focused = false;
        toggling_cursor_focus = false;
    }
}

bool RWindow::SetCursorPosition(const std::pair<int, int>& position) {
    if (!IsFocused())
        return false;

    if (!IsVisible())
        return false;

    if (position.first < 0 || position.first > GetWidth())
        return false;

    if (position.second < 0 || position.second > GetHeight())
        return false;

    POINT screen_pos = { position.first, position.second };
    if (!ClientToScreen(platform->hwnd, &screen_pos))
        return false;

    if (!SetCursorPos(screen_pos.x, screen_pos.y))
        return false;

    return true;
}

std::pair<int, int> RWindow::SetCursorCenter() {
    auto current_cursor_pos = GetAccurateCursorCoordinates();
    RECT client_rect;
    GetClientRect(platform->hwnd, &client_rect);
    POINT center((client_rect.right - client_rect.left) / 2, (client_rect.bottom - client_rect.top) / 2 );
    ClientToScreen(platform->hwnd, &center);
    SetCursorPos(center.x, center.y);

    return current_cursor_pos;
}

void RWindow::PollEvents() {
    MSG msg;
    while (PeekMessage(&msg, nullptr, 0, 0, PM_REMOVE))
        TranslateMessage(&msg),
        DispatchMessage(&msg);
}

LRESULT CALLBACK RWindow::Platform::WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
    auto* window = reinterpret_cast<RWindow*>( GetWindowLongPtr(hwnd, GWLP_USERDATA) );
    switch (uMsg) {
            case WM_CLOSE: {
                window->processOnClose();
                DestroyWindow(hwnd);
            }

            case WM_DESTROY: {
                exit(0);
            }

            case WM_SIZE: {
                window->SetSizeWithoutEvent({ LOWORD(lParam), HIWORD(lParam) });
                auto eventData = WindowResizeRequestEvent();
                eventData.Size = { window->GetWidth(), window->GetHeight() };
                // TODO: Implement eWindow->processOnResize()
                window->OnResizeRequest(eventData);
                window->OnResizeRequestEvent(eventData);
                break;
            }

            case WM_SETFOCUS: {
                window->processFocusIn();
                window->focused = true;

                if (window->GetCursorFocused())
                    window->toggling_cursor_focus = true;

                // Cancels window flashing.
                // TODO check if we're flashing before this.
                static FLASHWINFO flash_inter {sizeof(FLASHWINFO), hwnd, FLASHW_STOP, 0, 0};
                FlashWindowEx(&flash_inter);
                break;
            }

            case WM_KILLFOCUS: {
                window->processFocusOut();

                if (window->GetCursorFocused()) {
                    ClipCursor(nullptr);
                    window->toggling_cursor_focus = true;
                }
                window->focused = false;
            }

            case WM_SETCURSOR: {
                if (LOWORD(lParam) == HTCLIENT && window->GetCursorVisible() == false)
                    SetCursor(nullptr);
                break;
            }

        case WM_KEYDOWN: {
                auto key = GetKeyFromWindowsScancode((WindowsScancode) wParam);

                if (window->IsKeyRepeat())
                    window->processKeyPress(key);
                else if ((lParam & 1 << 30) == 0)
                        window->processKeyPress(key);
                break;
        }

            case WM_KEYUP: {
                auto key = GetKeyFromWindowsScancode((WindowsScancode) wParam);
                window->processKeyRelease(key);
                break;
            }

            // Mouse Buttons.
            case WM_MOUSEWHEEL: {
                int wheel_delta = GET_WHEEL_DELTA_WPARAM(wParam);
                // Clamp to normalized range [+1 or -1].
                // TODO: Test this and make sure sign is the same on both platforms for each direction.
                wheel_delta = (wheel_delta > 0) ? 1 : -1;
                // TODO: Determine sign of wheel_delta for each direction, (and on linux too), and document this.
                window->processMouseWheel(wheel_delta);
                break;
            }

            case WM_LBUTTONDOWN: {
                window->processMousePress(MouseButtons::Left);
                break;
            }

            case WM_LBUTTONUP: {
                window->processMouseRelease(MouseButtons::Left);
                break;
            }

            case WM_RBUTTONDOWN: {
                window->processMousePress(MouseButtons::Right);
                break;
            }

            case WM_RBUTTONUP: {
                window->processMouseRelease(MouseButtons::Right);
                break;
            }

            case WM_MBUTTONDOWN: {
                window->processMousePress(MouseButtons::Middle);
                break;
            }

            case WM_MBUTTONUP: {
                window->processMouseRelease(MouseButtons::Middle);
                break;
            }

            case WM_XBUTTONDOWN: {
                WORD button = GET_XBUTTON_WPARAM(wParam);
                if (button == XBUTTON1)
                    window->processMousePress(MouseButtons::Mouse4);

                if (button == XBUTTON2)
                    window->processMousePress(MouseButtons::Mouse5);
                break;
            }

            case WM_XBUTTONUP: {
                WORD button = GET_XBUTTON_WPARAM(wParam);
                if (button == XBUTTON1)
                    window->processMouseRelease(MouseButtons::Mouse4);

                if (button == XBUTTON2)
                    window->processMouseRelease(MouseButtons::Mouse5);
                break;
        }

        //This is the same as "Motion Notify" in the X Window System.
        case WM_MOUSEMOVE: {
                if (window->toggling_cursor_focus) {
                    RECT rect;
                    if (GetClientRect(hwnd, &rect)) {
                        POINT top_left = { rect.left, rect.top };
                        POINT bottom_right = { rect.right, rect.bottom };
                        MapWindowPoints(hwnd, nullptr, &top_left, 1);
                        MapWindowPoints(hwnd, nullptr, &bottom_right, 1);

                        RECT clip_rect = { top_left.x, top_left.y, bottom_right.x, bottom_right.y };
                        if (ClipCursor(&clip_rect))
                            window->toggling_cursor_focus = false;
                    }
                }
        }
        break;
    }
    return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

bool fullscreenmode = false;
bool open = false;

void RWindow::Raise() { SetWindowPos(platform->hwnd, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE); }
void RWindow::Lower() { SetWindowPos(platform->hwnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE); }

//TODO somehow check if our window is entirely obscured by other windows.
bool RWindow::IsVisible() const {
    if (!IsWindow(platform->hwnd))
        return false;
    if (IsIconic(platform->hwnd))
        return false;
    return true;
}

void RWindow::Flash() {
    if (!focused) {
        FLASHWINFO fi;
        fi.cbSize = sizeof(FLASHWINFO);
        fi.hwnd = platform->hwnd;
        fi.dwFlags = FLASHW_ALL;
        fi.uCount = 0;
        fi.dwTimeout = 0;
        FlashWindowEx(&fi);
    }
}

void RWindow::DestroyOSWindowHandle() {
    DestroyWindow(platform->hwnd);
}

void RWindow::DisableResizing() {
    RECT rect;
    GetWindowRect(platform->hwnd, &rect);

    LONG style = GetWindowLong(platform->hwnd, GWL_STYLE);
    style &= ~(WS_THICKFRAME | WS_MAXIMIZEBOX);

    SetWindowLong(platform->hwnd, GWL_STYLE, style);
    SetWindowPos(platform->hwnd, nullptr, rect.left, rect.top, rect.right - rect.left,
    rect.bottom - rect.top, SWP_FRAMECHANGED | SWP_NOZORDER | SWP_NOOWNERZORDER);

    this->resizable = false;
}

std::pair<int, int> RWindow::GetAccurateCursorCoordinates() const {
    POINT point;
    GetCursorPos(&point);
    ScreenToClient(platform->hwnd, &point);
    return { point.x, point.y };
}

void RWindow::SetCursorVisible(bool cursor_enable) {
    cursor_visible = cursor_enable;
}

bool RWindow::GetCursorVisible() {
    return cursor_visible;
}

void RWindow::SetKeyRepeatEnabled(bool state) {
    key_repeat = state;
}

std::pair<int, int> RWindow::GetSize() const {
    RECT rect;
    GetClientRect(platform->hwnd, &rect);
    return { (rect.right - rect.left), (rect.bottom - rect.top) };
}

std::pair<int, int> RWindow::GetPosition() const {
    RECT rect;
    if (GetWindowRect(platform->hwnd, &rect))
        return { rect.left, rect.top };

    return { -1, -1 };
}

void RWindow::SetPosition(int x, int y) {
    SetWindowPos(platform->hwnd, nullptr, x, y, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

void RWindow::SetPosition(const std::pair<int, int>& pos) {
    SetPosition(pos.first, pos.second);
}

void RWindow::Fullscreen() {
    platform->window_position_before_fullscreen = GetPosition();
    platform->window_size_before_fullscreen = GetSize();

    platform->window_position_before_fullscreen = this->GetPosition();
    SetWindowLong(platform->hwnd, GWL_STYLE, GetWindowLong(platform->hwnd, GWL_STYLE) & ~WS_OVERLAPPEDWINDOW);
    SetWindowPos(platform->hwnd, HWND_TOP, 0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN), SWP_FRAMECHANGED | SWP_NOOWNERZORDER);

    fullscreen_mode = true;
}

void RWindow::RestoreFromFullscreen() {
    SetWindowLong(platform->hwnd, GWL_STYLE, GetWindowLong(platform->hwnd, GWL_STYLE) | WS_OVERLAPPEDWINDOW);

    RECT rect = {0, 0, platform->window_size_before_fullscreen.first, platform->window_size_before_fullscreen.second };

    AdjustWindowRectEx(&rect, GetWindowLong(platform->hwnd, GWL_STYLE), FALSE, GetWindowLong(platform->hwnd, GWL_EXSTYLE));

    int window_width = rect.right - rect.left;
    int window_height = rect.bottom - rect.top;

    SetWindowPos(
        platform->hwnd,
        nullptr,
        platform->window_position_before_fullscreen.first,
        platform->window_position_before_fullscreen.second,
        window_width,
        window_height,
        SWP_FRAMECHANGED | SWP_NOZORDER | SWP_NOOWNERZORDER
    );

    DisableResizing();

    this->width = platform->window_position_before_fullscreen.first;
    this->height = platform->window_position_before_fullscreen.second;

    fullscreen_mode = false;
}

void RWindow::SetCursorStyle(CursorStyle style) const {}

RWindow::RWindow() {
    extant = this;
    platform = new Platform();
}

RWindow::RWindow(const std::string& wTitle, int wWidth, int wHeight,
bool wFullscreen, bool wResizable, bool wVsync) :
title(wTitle), width(wWidth), height(wHeight), fullscreen_mode(wFullscreen), resizable(wResizable),
vsync(wVsync) {
    extant = this;
    platform = new Platform();
}

void RWindow::DialogOK(const std::string& title, const std::string& message_box_text) {
    MessageBox(nullptr, message_box_text.c_str(), title.c_str(), MB_OK);
}

HGLRC glContext;
HMODULE gl_lib = nullptr;
bool constructor_success = true;

namespace OpenGL {
    typedef const GLubyte* (APIENTRY *GLGETSTRINGPROC)(GLenum name);
    GLGETSTRINGPROC glGetString = nullptr;

    typedef BOOL (WINAPI *WGLMAKECURRENTPROC)(HDC hdc, HGLRC hglrc);
    WGLMAKECURRENTPROC wglMakeCurrent = nullptr;

    typedef FARPROC (WINAPI *WGLGETPROCADDRPROC)(LPCSTR lpszProc);
    WGLGETPROCADDRPROC wglGetProcAddress = nullptr;


    typedef HGLRC (WINAPI *WGLCREATECONTEXTPROC)(HDC hdc);
    WGLCREATECONTEXTPROC wglCreateContext = nullptr;
}

bool OpenGLWindow::Open() {
    if (!constructor_success)
        return false;

  platform->hInstance = GetModuleHandle(nullptr);

  WNDCLASS wc = { };
  wc.lpfnWndProc = RWindow::Platform::WindowProc;
  wc.hInstance = platform->hInstance;
  wc.lpszClassName = "RWindowClass";
  RegisterClass(&wc);

    RECT rect = { 0, 0, width, height };
    AdjustWindowRectEx(&rect, WS_OVERLAPPEDWINDOW, FALSE, 0);

    int win_width = rect.right - rect.left;
    int win_height = rect.bottom - rect.top;

    platform->hwnd = CreateWindowEx(
        0,
        "RWindowClass",
        title.c_str(),
        WS_OVERLAPPEDWINDOW,
        CW_USEDEFAULT, CW_USEDEFAULT, win_width, win_height,
        nullptr,
        nullptr,
        platform->hInstance,
        nullptr
    );
    SetWindowLongPtr(platform->hwnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(this));
  PIXELFORMATDESCRIPTOR pfd = {
      sizeof(PIXELFORMATDESCRIPTOR),
      1,
      PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
      PFD_TYPE_RGBA,
      24,
      0, 0, 0, 0, 0, 0,
      0,
      0,
      0,
      0, 0, 0, 0,
      24,
      8,
      0,
      PFD_MAIN_PLANE,
      0,
      0, 0, 0
  };

  platform->hdc = GetDC(platform->hwnd);
  int pixelFormat = ChoosePixelFormat(platform->hdc, &pfd);
  SetPixelFormat(platform->hdc, pixelFormat, &pfd);
  glContext = OpenGL::wglCreateContext(platform->hdc);
  OpenGL::wglMakeCurrent(platform->hdc, glContext);
  ShowWindow(platform->hwnd, SW_SHOW);
  open = true;
    return true;
}

std::string OpenGLWindow::GetGraphicsDriverVendor() {
    return std::string(reinterpret_cast<const char*>(OpenGL::glGetString(GL_VENDOR)));
}

void OpenGLWindow::SwapBuffers() {
    ::SwapBuffers(platform->hdc);
}

void OpenGLWindow::SetVsyncEnabled(bool b) {
    typedef BOOL(WINAPI* PFNWGLSWAPINTERVALEXTPROC)(int interval);
    auto wglSwapIntervalEXT = (PFNWGLSWAPINTERVALEXTPROC) OpenGL::wglGetProcAddress("wglSwapIntervalEXT");

    if (wglSwapIntervalEXT)
        wglSwapIntervalEXT(b ? 1 : 0);
}

bool OpenGLWindow::SoftwareRendered() {
    std::string renderer(reinterpret_cast<const char*>(OpenGL::glGetString(GL_RENDERER)));
    if (renderer.find("llvmpipe"))
        return true;
    if (renderer.find("softpipe"))
        return true;
    if (renderer.find("GDI Generic"))
        return true;
    if (GetGraphicsDriverVendor().find("Mesa"))
        return true;
    return false;
}

OpenGLWindow::OpenGLWindow(const std::string& title, int width, int height, uint8_t gl_major, uint8_t gl_minor)
: gl_major(gl_major), gl_minor(gl_minor), RWindow(title, width, height) {
    gl_lib = LoadLibrary("opengl32.dll");

    if (!gl_lib)
        Logger::Error("opengl32.dll couldn't be found. If you have it, you can try putting it in the same directory as the executable."),
        constructor_success = false;

    OpenGL::glGetString = (OpenGL::GLGETSTRINGPROC) GetProcAddress(gl_lib, "glGetString");
    OpenGL::wglMakeCurrent = (OpenGL::WGLMAKECURRENTPROC) GetProcAddress(gl_lib, "wglMakeCurrent");
    OpenGL::wglGetProcAddress = (OpenGL::WGLGETPROCADDRPROC) GetProcAddress(gl_lib, "wglGetProcAddress");
    OpenGL::wglCreateContext = (OpenGL::WGLCREATECONTEXTPROC) GetProcAddress(gl_lib, "wglCreateContext");
}
